/**
 * @file gaim-galago.c Gaim Galago plugin
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#define GAIM_PLUGINS

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "account.h"
#include "blist.h"
#include "debug.h"
#include "plugin.h"
#include "server.h"
#include "version.h"

#include <libgalago/galago.h>
#include <dbus/dbus-glib.h>
#include <ctype.h>
#include <string.h>

#ifdef ENABLE_NLS
#  include <locale.h>
#  include <libintl.h>
#  define _(x) gettext(x)
#  ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#  else
#    define N_(String) (String)
#  endif
#else
#  include <locale.h>
#  define N_(String) (String)
#  define _(x) (x)
#  define ngettext(Singular, Plural, Number) ((Number == 1) ? (Singular) : (Plural))
#endif

static GHashTable *person_table = NULL;
static GalagoPerson *me = NULL;

static const char *
get_service_for_prpl_id(const char *prpl_id, const char *user)
{
	if (prpl_id == NULL)
		return NULL;

	if (!g_ascii_strcasecmp(prpl_id, "prpl-oscar"))
	{
		if (isdigit(*user))
			return GALAGO_SERVICE_ID_ICQ;
		else
			return GALAGO_SERVICE_ID_AIM;
	}
	else if (!g_ascii_strcasecmp(prpl_id, "prpl-novell"))
		return GALAGO_SERVICE_ID_GROUPWISE;
	else if (!g_ascii_strcasecmp(prpl_id, "prpl-gg"))
		return GALAGO_SERVICE_ID_GADUGADU;

	if (!strncmp(prpl_id, "prpl-", 5))
		return prpl_id + 5;

	return prpl_id;
}

static GalagoService *
get_galago_service(GaimAccount *account)
{
	const char *username, *protocol_id, *service_id;

	g_return_val_if_fail(account != NULL, NULL);

	username    = gaim_account_get_username(account);
	protocol_id = gaim_account_get_protocol_id(account);

	service_id = get_service_for_prpl_id(protocol_id, username);

	return galago_service_new(service_id, service_id, TRUE, 0);
}

static GalagoAccount *
get_my_galago_account(GaimAccount *account)
{
	GalagoService *service;
	const char *username;

	service = get_galago_service(account);

	if (service == NULL)
		return NULL;

	username = gaim_account_get_username(account);

	return galago_account_new(service, me, username);
}

static const char *
generate_person_id(GaimContact *contact)
{
	static int id = 0;
	static char temp[64];

	g_return_val_if_fail(contact != NULL, NULL);

	g_snprintf(temp, sizeof(temp), "person-%d", id++);

	return temp;
}

static void
update_avatar(GaimBuddy *buddy, GalagoAccount *gaccount)
{
	GaimBuddyIcon *buddy_icon;
	GalagoAvatar *avatar;
	const unsigned char *data;
	unsigned char *old_data = NULL;
	size_t len, old_len = 0;

	if ((buddy_icon = gaim_buddy_get_icon(buddy)) == NULL)
		return;

	data   = gaim_buddy_icon_get_data(buddy_icon, &len);
	avatar = galago_account_get_avatar(gaccount, FALSE);

	if (avatar != NULL)
		galago_avatar_get_image_data(avatar, &old_data, &old_len);

	if (avatar == NULL || len != old_len || memcmp(data, old_data, len))
	{
		GalagoService *service = galago_account_get_service(gaccount);

		gaim_debug_info("galago",
						"Setting avatar for %s on account %s on %s\n",
						buddy->name, gaim_account_get_username(buddy->account),
						galago_service_get_id(service));
		galago_avatar_new(gaccount, data, len);
	}
}

static void
buddy_idle_changed_cb(GaimBuddy *buddy)
{
	GalagoPresence *presence;
	GalagoService *service;
	GalagoAccount *my_gaccount, *gaccount;
	dbus_uint32_t idle_time = 0;
	gboolean idle;

	my_gaccount = get_my_galago_account(buddy->account);
	service     = galago_account_get_service(my_gaccount);

	gaccount = galago_service_get_account(service, buddy->name, FALSE);

	if (gaccount == NULL)
		return;

	/* This is a hack, but a necessary one. */
	update_avatar(buddy, gaccount);

	presence = galago_account_get_presence(gaccount, FALSE);

	if (presence == NULL)
		return;

	idle = (buddy->idle > 0);

	if (idle)
		idle_time = time(NULL) - buddy->idle;

	galago_presence_set_idle(presence, idle, idle_time);
}

static void
buddy_status_changed_cb(GaimBuddy *buddy)
{
	GaimPlugin *prpl;
	GaimPluginProtocolInfo *prpl_info = NULL;
	GalagoPresence *presence;
	GalagoService *service;
	GalagoAccount *my_gaccount, *gaccount;
	GalagoStatus *status;
	GalagoStatusType status_type;
	const char *status_id;
	char *status_message = NULL;

	my_gaccount = get_my_galago_account(buddy->account);
	service     = galago_account_get_service(my_gaccount);

	gaccount = galago_service_get_account(service, buddy->name, FALSE);

	if (gaim_account_is_connected(buddy->account))
	{
		if (gaccount == NULL)
		{
			GalagoPerson *person;
			GaimContact *contact;

			contact = (GaimContact *)((GaimBlistNode *)buddy)->parent;

			person = g_hash_table_lookup(person_table, contact);

			if (person == NULL)
			{
				person = galago_person_new(generate_person_id(contact), TRUE);
				g_hash_table_insert(person_table, contact, person);
			}

			gaccount = galago_account_new(service, person, buddy->name);
			galago_account_set_display_name(gaccount,
											gaim_buddy_get_alias(buddy));
			galago_account_add_contact(my_gaccount, gaccount);
		}
	}
	else
	{
		if (gaccount != NULL)
			galago_account_set_presence(gaccount, NULL);

		return;
	}

	if (!GAIM_BUDDY_IS_ONLINE(buddy))
	{
		galago_account_set_connected(gaccount, FALSE);

		return;
	}

	if (!galago_account_is_connected(gaccount))
		galago_account_set_connected(gaccount, TRUE);

	presence = galago_presence_new(gaccount);

	gaim_debug_info("galago",
					"Adding presence for %s on account %s on %s\n",
					buddy->name, gaim_account_get_username(buddy->account),
					galago_service_get_id(service));

	/* This will all change when gaim's new status API goes in. */
	if (buddy->uc & UC_UNAVAILABLE)
	{
		status_type = GALAGO_STATUS_AWAY;
		status_id   = GALAGO_STATUS_ID_AWAY;
	}
	else
	{
		status_type = GALAGO_STATUS_AVAILABLE;
		status_id   = GALAGO_STATUS_ID_AVAILABLE;
	}

	prpl = gaim_find_prpl(gaim_account_get_protocol_id(buddy->account));

	if (prpl != NULL)
		prpl_info = GAIM_PLUGIN_PROTOCOL_INFO(prpl);

	if (prpl != NULL && prpl_info->status_text != NULL)
	{
		char *temp = prpl_info->status_text(buddy);

		if (temp != NULL)
		{
			status_message = g_strdup(temp);

			g_free(temp);
		}
	}

	if (galago_presence_has_status(presence, status_id))
	{
		status = galago_presence_get_status(presence, status_id);

		if (status_message != NULL)
			galago_status_set_attr_string(status, "message", status_message);
	}
	else
	{
		status = galago_status_new(status_type, status_id, NULL, TRUE);

		if (status_message != NULL)
			galago_status_set_attr_string(status, "message", status_message);

		galago_presence_add_status(presence, status);
	}

	if (status_message != NULL)
		g_free(status_message);

	update_avatar(buddy, gaccount);
}

static void
signed_on_cb(GaimConnection *gc)
{
	GaimBlistNode *bnode, *cnode, *group;
	GaimAccount *account;
	GalagoPerson *me;
	GalagoAccount *my_gaccount;

	me = galago_core_get_me(TRUE, FALSE);

	account     = gaim_connection_get_account(gc);
	my_gaccount = get_my_galago_account(account);

	galago_account_set_connected(my_gaccount, TRUE);

	for (group = gaim_get_blist()->root; group != NULL; group = group->next)
	{
		for (cnode = group->child; cnode != NULL; cnode = cnode->next)
		{
			if (GAIM_BLIST_NODE_IS_CONTACT(cnode))
			{
				for (bnode = cnode->child; bnode != NULL; bnode = bnode->next)
				{
					GaimBuddy *buddy = (GaimBuddy *)bnode;

					if (account == buddy->account)
					{
						buddy_status_changed_cb(buddy);
						buddy_idle_changed_cb(buddy);
					}
				}
			}
		}
	}
}

static void
signed_off_cb(GaimConnection *gc)
{
	GaimAccount *account;
	GalagoAccount *my_gaccount;
	GalagoPerson *me;
	GalagoPresence *presence;

	me = galago_core_get_me(TRUE, FALSE);

	account     = gaim_connection_get_account(gc);
	my_gaccount = get_my_galago_account(account);

	galago_account_set_connected(my_gaccount, FALSE);

	presence = galago_presence_new(my_gaccount);

	if (!galago_presence_has_status(presence, GALAGO_STATUS_ID_OFFLINE))
	{
		GalagoStatus *status;

		status = galago_status_new(GALAGO_STATUS_OFFLINE,
								   GALAGO_STATUS_ID_OFFLINE, NULL, TRUE);
		galago_presence_add_status(presence, status);
	}
}

static void
setup_accounts(void)
{
	GList *l;

	me = galago_person_me_new(TRUE);

	for (l = gaim_accounts_get_all(); l != NULL; l = l->next)
	{
		GaimAccount *account = (GaimAccount *)l->data;
		GalagoAccount *my_gaccount;

		my_gaccount = get_my_galago_account(account);

		galago_account_set_connected(my_gaccount,
									 gaim_account_is_connected(account));
	}
}

static void
scan_presences(void)
{
	GaimBlistNode *bnode, *cnode, *group;

	if (gaim_get_blist() == NULL)
		return;

	for (group = gaim_get_blist()->root; group != NULL; group = group->next)
	{
		for (cnode = group->child; cnode != NULL; cnode = cnode->next)
		{
			if (GAIM_BLIST_NODE_IS_CONTACT(cnode))
			{
				for (bnode = cnode->child; bnode != NULL; bnode = bnode->next)
				{
					GaimBuddy *buddy = (GaimBuddy *)bnode;

					if (gaim_account_is_connected(buddy->account))
						buddy_status_changed_cb(buddy);
				}
			}
		}
	}
}

static void
log_handler(GalagoLogLevel level, const char *msg, va_list args,
			void *user_data)
{
	GaimDebugLevel debug_level = (GaimDebugLevel)level;

	if (level == GALAGO_LOG_ASSERT)
		debug_level = GAIM_DEBUG_WARNING;

	gaim_debug_vargs(debug_level, "libgalago", msg, args);
}

static gboolean
plugin_load(GaimPlugin *plugin)
{
	galago_log_set_handler(GALAGO_LOG_ALL, log_handler, NULL);

	if (!galago_glib_init("gaim", TRUE, NULL))
		return FALSE;

	if (!galago_is_connected())
	{
		gaim_debug(GAIM_DEBUG_ERROR, "gaim-galago",
				   "Unable to connect to Galago service\n");

		return FALSE;
	}

	if (!galago_is_registered())
	{
		gaim_debug(GAIM_DEBUG_ERROR, "gaim-galago",
				   "Unable to register with the Galago service\n");

		return FALSE;
	}

	person_table = g_hash_table_new(NULL, NULL);

	/* Connect the signals */
	gaim_signal_connect(gaim_blist_get_handle(), "buddy-away",
						plugin, GAIM_CALLBACK(buddy_status_changed_cb), NULL);
	gaim_signal_connect(gaim_blist_get_handle(), "buddy-back",
						plugin, GAIM_CALLBACK(buddy_status_changed_cb), NULL);
	gaim_signal_connect(gaim_blist_get_handle(), "buddy-idle",
						plugin, GAIM_CALLBACK(buddy_idle_changed_cb), NULL);
	gaim_signal_connect(gaim_blist_get_handle(), "buddy-unidle",
						plugin, GAIM_CALLBACK(buddy_idle_changed_cb), NULL);
	gaim_signal_connect(gaim_blist_get_handle(), "buddy-idle-updated",
						plugin, GAIM_CALLBACK(buddy_idle_changed_cb), NULL);
	gaim_signal_connect(gaim_blist_get_handle(), "buddy-signed-on",
						plugin, GAIM_CALLBACK(buddy_status_changed_cb), NULL);
	gaim_signal_connect(gaim_blist_get_handle(), "buddy-signed-off",
						plugin, GAIM_CALLBACK(buddy_status_changed_cb), NULL);

	gaim_signal_connect(gaim_connections_get_handle(), "signed-on",
						plugin, GAIM_CALLBACK(signed_on_cb), NULL);
	gaim_signal_connect(gaim_connections_get_handle(), "signed-off",
						plugin, GAIM_CALLBACK(signed_off_cb), NULL);

	setup_accounts();
	scan_presences();

	return TRUE;
}

static gboolean
plugin_unload(GaimPlugin *plugin)
{
	galago_uninit();

	g_hash_table_destroy(person_table);
	person_table = NULL;

	return TRUE;
}

static GaimPluginInfo info =
{
	GAIM_PLUGIN_MAGIC,
	GAIM_MAJOR_VERSION,
	GAIM_MINOR_VERSION,
	GAIM_PLUGIN_STANDARD,                             /**< type           */
	NULL,                                             /**< ui_requirement */
	0,                                                /**< flags          */
	NULL,                                             /**< dependencies   */
	GAIM_PRIORITY_DEFAULT,                            /**< priority       */

	"core-chipx86-galago",                            /**< id             */
	"Galago",                                         /**< name           */
	VERSION,                                          /**< version        */
	                                                  /**  summary        */
	N_("Turns gaim into a feed for Galago."),
	                                                  /**  description    */
	N_("Turns gaim into a feed for Galago."),
	"Christian Hammmond <chipx86@gnupdate.org>",      /**< author         */
	"http://galago.sourceforge.net/",                 /**< homepage       */

	plugin_load,                                      /**< load           */
	plugin_unload,                                    /**< unload         */
	NULL,                                             /**< destroy        */

	NULL,                                             /**< ui_info        */
	NULL                                              /**< extra_info     */
};

static void
init_plugin(GaimPlugin *plugin)
{
}

GAIM_INIT_PLUGIN(galago, init_plugin, info)
